@@ -1,5 +1,6 @@
 """Meta-programming tools to build declarative frameworks"""
 from i2.deco import (
+    FuncFactory,
     preprocess,
     postprocess,
     preprocess_arguments,
@@ -48,7 +49,9 @@ from i2.wrapper import (
 )
 
 from i2.util import (
-    Literal,
+    copy_func,
+    get_app_data_folder,
+    LiteralVal,
     path_extractor,
     get_function_body,
     lazyprop,
@@ -57,3 +60,7 @@ from i2.util import (
     mk_sentinel,
     ensure_identifiers,
 )
+
+from i2.footprints import MethodTrace
+
+from i2.itypes import validate_literal
