@@ -2,11 +2,13 @@
 import inspect
 from collections import defaultdict
 from contextlib import suppress
-from functools import wraps, partial
+from functools import partial
+from functools import wraps  # TODO: Overwritten. Remove?
 from inspect import Signature, signature, Parameter
 from itertools import chain
 from typing import Callable, Tuple, Dict, Any, TypeVar
 
+from i2.util import mk_sentinel
 from i2.signatures import Sig, kind_forgiving_func, name_of_obj
 
 # keep the imports below here because might be referenced (or take care of refs)
@@ -28,11 +30,30 @@ def identity(obj: T) -> T:
 from functools import WRAPPER_ASSIGNMENTS, WRAPPER_UPDATES, update_wrapper
 
 
+# Overwrites the imported functools.wraps
 def wraps(wrapped, assigned=WRAPPER_ASSIGNMENTS, updated=WRAPPER_UPDATES):
     return partial(update_wrapper, wrapped=wrapped, assigned=assigned, updated=updated)
 
 
+def _resolve_inclusion(include, exclude, super_set):
+    if isinstance(include, str):
+        include = include.split()
+    if isinstance(exclude, str):
+        exclude = exclude.split()
+    if exclude is None:
+        exclude = set()
+    include = [x for x in (include or super_set) if x not in exclude]
+    return include
+
+
+def _not_set_repr(self):
+    return 'NotSet'
+
+
+NotSet = mk_sentinel('NotSet', repr_=_not_set_repr)
+
 # ---------------------------------------------------------------------------------------
+# TODO: https://github.com/i2mint/i2/issues/48
 class FuncFactory:
     """Make a function factory.
 
@@ -64,7 +85,7 @@ class FuncFactory:
 
     >>> factory = FuncFactory(foo)
     >>> factory
-    <FuncFactory(foo)>(b, *, c=2) -> Callable[..., float]
+    <FuncFactory(foo)>(a, b, *, c=2) -> Callable[..., float]
 
     (Note that the repr even reuses ``foo``'s return annotation to tell us that our
     factory will return a callable that returns that type (if the annotation is a type).
@@ -76,6 +97,22 @@ class FuncFactory:
     >>> f(10)
     23
 
+    Note that:
+
+    >>> ff = factory(2, 3)  # equivalent to ``factory(a=2, b=3)``
+    >>> ff(c=10)
+    16
+
+    Further, you can tell ``FuncFactory`` to ``include`` or ``exclude`` specific
+    arguments, using their names or indices to specify them.
+
+    >>> factory_no_a = FuncFactory(foo, exclude=['a'])
+    >>> factory_no_a
+    <FuncFactory(foo)>(b, *, c=2) -> Callable[..., float]
+    >>> g = factory_no_a(2, 3)  # equivalent to ``factory(b=2, c=3)`` as no ``a`` here
+    >>> g(10)
+    23
+
     Recipe: Say you're normalizing some data accessor into callback functions and you
     want to create functions that provide a specific object when called (with no args).
     Sure, you can do this by specifying ``lambda: obj`` every time, but lambdas can be
@@ -100,22 +137,59 @@ class FuncFactory:
 
     """
 
-    def __init__(self, func):
+    def __init__(self, func, *, include=(), exclude=()):
         self.func = func
+        func_sig = Sig(func)
+        include, exclude = map(func_sig.get_names, (include, exclude))
+        include = _resolve_inclusion(include, exclude, func_sig.names)
+        self.include = include
 
-        sig = Sig(func)
-        factory_sig = Sig(sig, return_annotation=Callable[..., Any]) - sig.names[0]
-        if sig.return_annotation is not sig.empty:
+        actual_factory_sig = Sig(func_sig, return_annotation=Callable[..., partial])
+        actual_factory_sig = actual_factory_sig[self.include]
+        # previous I did the following, but don't know why:
+        # factory_sig = factory_sig - sig.names[0]
+        # Now I know: Because in the func(obj, **params) case it makes more sense
+        # that the sig be **params
+        # This led to extend to the include/exclude functionality.
+        if func_sig.return_annotation is not func_sig.empty:
             try:
-                factory_sig = Sig(
-                    factory_sig, return_annotation=Callable[..., sig.return_annotation]
+                actual_factory_sig = Sig(
+                    actual_factory_sig,
+                    return_annotation=Callable[..., func_sig.return_annotation],
                 )
             except TypeError:
                 pass
 
-        self.__signature__ = factory_sig
+        self.func_sig = func_sig
+        self.factory_sig = actual_factory_sig
+
+        self.__signature__ = actual_factory_sig  # TODO: Delete when #48 solved
+        # TODO: Uncomment below to resolved https://github.com/i2mint/i2/issues/48)
+        ## Add NotSet default to all non-defaulted params:
+        ## (To see why, go to See https://github.com/i2mint/i2/issues/48)
+        # shown_factory_sig = actual_factory_sig.ch_defaults(
+        #     **{name: NotSet for name in actual_factory_sig.required_names}
+        # )
+        # shown_factory_sig = shown_factory_sig[self.include]
+        # self.__signature__ = shown_factory_sig
+
+    @classmethod
+    def wrap(cls, include=(), exclude=()):
+        return partial(cls, include=include, exclude=exclude)
+
+    def _process_args_and_kwargs(self, args, kwargs):
+        _kwargs = self.factory_sig.kwargs_from_args_and_kwargs(
+            args, kwargs, allow_partial=True, ignore_kind=True
+        )
+        # Uncomment below to resolved https://github.com/i2mint/i2/issues/48)
+        # _kwargs = {k: v for k, v in _kwargs.items() if v is not NotSet}
+        __args, __kwargs = self.func_sig.args_and_kwargs_from_kwargs(
+            _kwargs, allow_partial=True, ignore_kind=False
+        )
+        return __args, __kwargs
 
     def __call__(self, *args, **kwargs):
+        args, kwargs = self._process_args_and_kwargs(args, kwargs)
         return partial(self.func, *args, **kwargs)
 
     def __repr__(self):
@@ -125,6 +199,13 @@ class FuncFactory:
     def func_returning_obj(cls, obj):
         return cls(identity)(obj)
 
+    def to_jdict(self):
+        return {'func': self.func}
+
+    @classmethod
+    def from_jdict(cls, jdict):
+        return FuncFactory(**jdict)
+
 
 def _double_up_as_factory(wrapped=None, *args, __decorator_func=None, **kwargs):
     """Util for double_up_as_factory, ``__decorator_func`` to be partialized"""
