@@ -1,6 +1,74 @@
 """Flattening maps and manipulating key paths"""
 
-from functools import reduce, wraps, partial
+from operator import attrgetter
+import importlib
+from itertools import chain
+
+
+def obj_to_str_path(
+    obj,
+    *,
+    sep='.',
+    name_of_obj=attrgetter('__qualname__'),
+    path_of_module=attrgetter('__module__'),
+):
+    """Get the dotpath reference for an object
+
+    >>> from inspect import Signature
+    >>> obj_to_str_path(Signature.replace)
+    'inspect.Signature.replace'
+
+    ``obj_to_str_path`` is the inverse of ``str_path_to_obj``
+
+    >>> assert str_path_to_obj(obj_to_str_path(Signature.replace)) == Signature.replace
+
+    Let's try with a different separator.
+
+    >>> path = obj_to_str_path(Signature.replace, sep='/')
+    >>> path
+    'inspect/Signature.replace'
+
+    Remember to specify the same ``sep`` when you do the inverse!
+
+    >>> assert str_path_to_obj(path, sep='/') == Signature.replace
+
+    You can also pass in your own ``name_of_obj`` and ``path_of_module`` functions.
+    For example you want a more permissive version of `name_of_obj` you may consider
+    ``i2.name_of_obj``.
+    Note, thought, that ``str_path_to_obj`` might not work as an inverse for
+    custom ``name_of_obj`` and ``path_of_module`` functions.
+    You may have to write your own inverse function in this case.
+
+    """
+    return sep.join((path_of_module(obj), name_of_obj(obj)))
+
+
+def str_path_to_obj(str_path: str, *, sep='.'):
+    """Loads and returns the object referenced by the string DOTPATH_TO_MODULE.OBJ_NAME
+    """
+
+    path_parts = iter(str_path.split(sep))
+    path_parts = chain.from_iterable(map(lambda x: x.split('.'), path_parts))
+    module_name = next(path_parts)
+    obj = importlib.import_module(module_name)  # assume it's a module
+    for item in path_parts:
+        new_module_name = module_name + '.' + item
+        try:
+            obj = importlib.import_module(new_module_name)
+            module_name = new_module_name
+        except ModuleNotFoundError:
+            while item:
+                try:
+                    obj = getattr(obj, item)
+                    item = next(path_parts, None)
+                except AttributeError:
+                    break
+    return obj
+
+
+# --------------------------------------------------------------------------------------
+
+from functools import reduce, wraps
 import operator
 from collections.abc import MutableMapping
 
@@ -586,7 +654,6 @@ class KeyPathTrans:
             key_paths = [k for k, v in key_paths.items() if v]
 
         for key_path in key_paths:
-
             if isinstance(key_path, str):
                 field = key_path
                 key_path = key_path.split(self.sep)
