@@ -1,19 +1,22 @@
-"""A few fundamental tools to operate on a fixed pool of objects (e.g. functions).
+"""A few fundamental tools to operate on a fixed collection of objects (e.g. functions).
 
 For functions you have:
 
-- `Pipe`: To compose functions (output of one fed as the input of the next)
-- `FuncFanout`: To apply multiple functions to the same inputs
-- `FlexFuncFanout`: Like `FuncFanout` but where the application of inputs is flexible.
-    That is, the functions "draw" their inputs from the a common pool, but don't choke
-    if there are extra unrecognized arguments.
-- `ParallelFuncs`: To make a dict-to-dict function, applying a specific function for each
-    input key (putting the result in that key in the output.
+- ``Pipe``: To compose functions (output of one fed as the input of the next)
+
+- ``FuncFanout``: To apply multiple functions to the same inputs
+
+- ``FlexFuncFanout``: Like `FuncFanout` but where the application of inputs is flexible.
+That is, the functions "draw" their inputs from the a common pool, but don't choke
+if there are extra unrecognized arguments.
+
+- ``ParallelFuncs``: To make a dict-to-dict function, applying a specific function for
+each input key (putting the result in that key in the output.
 
 For context managers you have:
 
-- `ContextFanout`: To hold multiple context managers as one (entering and exiting
-    together)
+- ``ContextFanout``: To hold multiple context managers as one (entering and exiting
+together)
 
 ![image](https://user-images.githubusercontent.com/1906276/138004878-bfe17115-c25f-4d22-9740-0fef983507c0.png)
 
@@ -22,10 +25,19 @@ from typing import Mapping, Iterable, Union, Callable, Any, TypeVar
 from inspect import Signature, signature
 
 
+# Note: For a more powerful, and customizable implementation,
+#  see `i2.signatures.name_of_obj`.
+# TODO: Analyse usages and reroute to i2.signatures.name_of_obj instead.
+#   Reason this duplicate exists might be because we wanted to keep multi_object
+#   completely independent from other modules (self-contained).
 def name_of_obj(o: object) -> Union[str, None]:
     """
     Tries to find the (or "a") name for an object, even if `__name__` doesn't exist.
 
+    This is a basic implementation, and it's not guaranteed to work for all objects.
+    For a more powerful, and customizable implementation,
+    see `i2.signatures.name_of_obj`.
+
     >>> name_of_obj(map)
     'map'
     >>> name_of_obj([1, 2, 3])
@@ -71,37 +83,90 @@ def uniquely_named_objects(
     objects: Iterable[Obj],
     exclude_names: Iterable[str] = (),
     obj_to_name: Callable[[Obj], str] = name_of_obj,
+    *,
+    name_for_position: dict = (),
 ):
     """Generate (name, object) pairs from an iterable of objects
 
     :param objects: Objects to be named
     :param exclude_names: Names that can't be used
     :param obj_to_name: Function that tries to get/make a name from an object
+    :param name_for_position: A ``{position_idx: name,...}`` mapping that instructs
+    ``uniquely_named_objects`` to use a specific name for a given position.
 
     >>> from functools import partial
     >>> objects = [map, [1], [1, 2], lambda x: x, partial(print, sep=",")]
     >>> g = uniquely_named_objects(objects)
     >>> names_and_objects = dict(g)
     >>> list(names_and_objects)
-    ['map', 'list', '_2', 'lambda_3', 'print']
-
-    That '_2' is there because both [1] and [1, 2] would be named `'list'`, so to avoid
-    that, a default name (revealing the position of the object in the input `objects`)
-    is given.
-
-    If we wanted to not allow the function to choose 'lambda_3' as a name, we can do
-    so with the `exclude_names` argument;
-
-    >>> list(dict(uniquely_named_objects(objects, exclude_names={'lambda_3'})))
     ['map', 'list', '_2', '_3', 'print']
 
+    That ``'_2'`` is there because both ``[1]`` and ``[1, 2]`` would be named `'list'`,
+    so to avoid that, a default name (revealing the position of the object in the
+    input `objects`) is given.
+    The '_3' comes from the fact that the ``lambda`` function doesn't have a proper
+    name (one that is a python identifier).
+
+    If we wanted the name for ``[1]`` to revert to the default positional name ``'_1'``,
+    we can achieve this by forbidding the name ``'list'``:
+
+    >>> list(dict(uniquely_named_objects(objects, exclude_names={'list'})))
+    ['map', '_1', '_2', '_3', 'print']
+
+    You could also acheive this by specifying this ``'_1'`` explicitly in the
+    ``name_for_position`` argument:
+
+    >>> list(dict(uniquely_named_objects(objects, name_for_position={1: '_1'})))
+    ['map', '_1', 'list', '_3', 'print']
+
+    The reason this `list` reappears as a name is that we didn't exclude it, and
+    the name is not taken by the ``[1]`` argument anymore.
+    To get the desired effect with ``name_for_position`` we could therefore do this:
+
+    >>> list(dict(uniquely_named_objects(objects, name_for_position={1: '_1', 2: '_2'})))
+    ['map', '_1', '_2', '_3', 'print']
+
+    Obviously, ``exclude_names`` is the right argument for the problem above, but
+    what ``name_for_position`` does give you is the ability to explicitly chose the
+    names you want to assign to all or some of the elements of your iterable.
+
+    >>> list(dict(uniquely_named_objects(
+    ...     objects, name_for_position={1: 'first_list', 2: 'second_list', 3: 'lambda'}))
+    ... )
+    ['map', 'first_list', 'second_list', 'lambda', 'print']
+
+    Extra notes:
+
+    See what ``uniquely_named_objects`` offers as parametrization:
+
+    - You can provide an exclusion list (though the handing of a conflict is hardcoded
+    and questionable)
+
+    - You can provide a ``obj_to_name`` function to control the naming of objects.
+    One trick to be aware of if objects have unique hashes: Make a
+    ``d = {obj: name,...}`` mapping and specify ``obj_to_name=d.get``.
+
+    - Any controllable way to decide on a name based on the position of the function
+    in the iterable (this could be useful!)
+
+    What you DO NOT have:
+
+    - Any way to choose names non-myopically: An object’s name cannot “see” the
+    objects around it to decide on a name (it can only see the names use by those behind
+    it through ``exclude_names``).
+
+    - Any “retries” or “alternative naming logic” if a chosen name conflicts with
+    ``exclude_names``
     """
     _exclude_names = set(exclude_names)
     for i, obj in enumerate(objects):
-        name = obj_to_name(obj)
-        if name == '<lambda>':
-            name = f'lambda_{i}'
-        if name is None or name in _exclude_names:
+        if i in name_for_position:
+            name = name_for_position[i]
+        else:
+            name = obj_to_name(obj)
+        # if name == '<lambda>':
+        # name = f'lambda_{i}'
+        if name is None or not name.isidentifier() or name in _exclude_names:
             name = f'_{i}'
             assert (
                 name not in _exclude_names
@@ -193,12 +258,27 @@ class MultiObj(Mapping):
     >>> mo = MultiObj({'this': [1], 'that': [1, 2]})
     >>> dict(mo)
     {'this': [1], 'that': [1, 2]}
+
+    You can specify an instance name and/or doc with the special (reserved) argument
+    names ``__name__`` and ``__doc__`` (which therefore can't be used as object names:
+
+    >>> mo = MultiObj(
+    ... this=[1], that=[1, 2], __name__='this_and_that', __doc__='Nothing much'
+    ... )
+    >>> dict(mo)
+    {'this': [1], 'that': [1, 2]}
+    >>> mo.__name__
+    'this_and_that'
+    >>> mo.__doc__
+    'Nothing much'
     """
 
     # Placing as staticmethod so that subclasses can overwrite if necessary
     auto_namer = staticmethod(uniquely_named_objects)
 
     def __init__(self, *unnamed, **named):
+        named = self._process_reserved_names(named)
+
         if len(unnamed) == 1 and isinstance(unnamed[0], Mapping) and len(named) == 0:
             # Special case where a single input is given: a mapping between names and obj
             self.objects = unnamed[0]
@@ -215,6 +295,14 @@ class MultiObj(Mapping):
             # Merge these
             self.objects = dict(named_unnamed, **named)
 
+    _reserved_names = ('__name__', '__doc__')
+
+    def _process_reserved_names(self, named_funcs):
+        for name in self._reserved_names:
+            if (value := named_funcs.pop(name, None)) is not None:
+                setattr(self, name, value)
+        return named_funcs
+
     def __iter__(self):
         yield from self.objects
 
@@ -234,6 +322,16 @@ class MultiObj(Mapping):
         else:
             raise AttributeError(f'Not an attribute: {item}')
 
+    def __repr__(self):
+        return f"<{type(self).__name__} containing: {', '.join(self)}>"
+
+    def __getstate__(self):
+        return self.__dict__
+
+    def __setstate__(self, state):
+        for k, v in state.items():
+            setattr(self, k, v)
+
 
 def iterable_of_callables_validation(funcs: Iterable[Callable]):
     if not isinstance(funcs, Iterable):
@@ -250,9 +348,13 @@ class MultiFunc(MultiObj):
         # The basic MultiObj initialization
         super().__init__(*unnamed_funcs, **named_funcs)
         # The extra stuff we want to do with functions
-        self.funcs = self.objects  # an alias, for better readability
         iterable_of_callables_validation(self.funcs.values())
 
+    @property
+    def funcs(self):
+        """Alias of .objects, for better readability in the context of MultiFunc"""
+        return self.objects
+
 
 def _signature_from_first_and_last_func(first_func, last_func):
     try:
@@ -266,8 +368,6 @@ def _signature_from_first_and_last_func(first_func, last_func):
     return Signature(input_params, return_annotation=return_annotation)
 
 
-# TODO: Give it a __name__ and make it more like a "normal" function so it works
-#  well when so assumed?
 class Pipe(MultiFunc):
     """Simple function composition. That is, gives you a callable that implements
 
@@ -305,6 +405,17 @@ class Pipe(MultiFunc):
     Traceback (most recent call last):
       ...
     ValueError: You need to specify at least one function!
+
+    You can specify an instance name and/or doc with the special (reserved) argument
+    names ``__name__`` and ``__doc__`` (which therefore can't be used as function names):
+
+    >>> f = Pipe(map, add_it=sum, __name__='map_and_sum', __doc__='Apply func and add')
+    >>> f(lambda x: x * 10, [1, 2, 3])
+    60
+    >>> f.__name__
+    'map_and_sum'
+    >>> f.__doc__
+    'Apply func and add'
     """
 
     def __init__(self, *unnamed_funcs, **named_funcs):
@@ -449,25 +560,32 @@ class FuncFanout(MultiFunc):
     >>> groot = lambda a: 'I am groot'
     >>> m = FuncFanout(foo, bar, groot)
     >>>
-    >>> m(3)
-    {'foo': 5, 'bar': 6, 'lambda_2': 'I am groot'}
+    >>> list(m(3))
+    [('foo', 5), ('bar', 6), ('_2', 'I am groot')]
+    >>> dict(m(3))
+    {'foo': 5, 'bar': 6, '_2': 'I am groot'}
 
-    Don't like that `lambda_2`?
+    Don't like that `_2` name?
     Well, If you specify names to the input functions, they'll be used instead of the
     ones found by the `MultObj.auto_namer`.
 
     >>> m = FuncFanout(foo, bar_results=bar, groot=groot)
-    >>> m(10)
+    >>> dict(m(10))
     {'foo': 12, 'bar_results': 20, 'groot': 'I am groot'}
 
-    `FuncFanout` uses the `call_generator` method to iterate through the functions,
-    yielding `(func_key, func_output)` pairs on the way.
-    The result of calling a `FuncFanout` is simply the gathering of those pairs in
-    a `dict`.
+    Or if you want your results as a tuple, you could do:
+
+    >>> tuple(dict(m(10)).values())
+    (12, 20, 'I am groot')
+
+    The above, gather in a ``dict`` is one way to get your data, but what calling a
+    ``FuncFanout`` instance actually gives you is a generator that yields the
+    ``(func_key, func_output)`` pairs one at a time
+
     Sometimes you may want/need more control though, and prefer to iterate through the
     pairs yourself, and in that case use `call_generator` directly.
 
-    >>> gen = m.call_generator(10)
+    >>> gen = m(10)
     >>> next(gen)
     ('foo', 12)
     >>> next(gen)
@@ -475,14 +593,21 @@ class FuncFanout(MultiFunc):
     >>> next(gen)
     ('groot', 'I am groot')
 
-    You know how you can do `dict(a=1, b=2)` or `dict({'a': 1, 'b': 2})` with dicts?
-    You can do that here too.
-    It's useful when the names (keys) of your functions aren't valid python names,
-    or not even strings.
+    So this gives you control on how you want your data.
+    Here's a recipe: Say you want to make a function that gives you the data as a dict
+    automatically. You can do this, using ``i2.Pipe``:
 
-    >>> m = FuncFanout({2: foo, "Bar results": bar, "I am groot": groot})
-    >>> m(10)
-    {2: 12, 'Bar results': 20, 'I am groot': 'I am groot'}
+    >>> f = Pipe(m, dict)
+    >>> f(10)
+    {'foo': 12, 'bar_results': 20, 'groot': 'I am groot'}
+
+    Or if you want a tuple:
+
+    >>> from operator import itemgetter, methodcaller
+    >>> from functools import partial
+    >>> f = Pipe(m, partial(map, itemgetter(1)), tuple)
+    >>> f(10)
+    (12, 20, 'I am groot')
 
     """
 
@@ -491,7 +616,7 @@ class FuncFanout(MultiFunc):
             yield name, func(*args, **kwargs)
 
     def __call__(self, *args, **kwargs):
-        return dict(self.call_generator(*args, **kwargs))
+        return self.call_generator(*args, **kwargs)
 
 
 from i2.signatures import ch_func_to_all_pk, tuple_the_args, Sig
