@@ -129,6 +129,44 @@ def test_attrs_used_by_method_computation():
     ) == {'x', 'a'}
 
 
+def test_object_dependencies():
+    from functools import cached_property
+    from i2.footprints import object_dependencies
+
+    class A:
+        def foo(self):
+            x = self.bar()
+            y = self.baz()
+            return x + y + self.prop + self.cached_prop
+
+        def bar(instance_name_could_be_anything):
+            return 2 * instance_name_could_be_anything.baz()
+
+        def baz(instance):
+            return 1
+
+        @cached_property
+        def cached_prop(self):
+            return 10 * self.baz()
+
+        @property
+        def prop(self):
+            return 100 * self.bar()
+
+        simple_prop = 1
+
+    expected = {
+        'bar': {'baz'},
+        'baz': set(),
+        'foo': {'bar', 'baz', 'prop', 'cached_prop'},
+        'cached_prop': {'baz'},
+        'prop': {'bar'},
+    }
+    assert (
+        object_dependencies(A) == expected
+    ), f'Expected:\n\t{expected}\n\nGot:\n\t{object_dependencies(A)=}'
+
+
 ## Order conserving
 # def test_attrs_used_by_method():
 #     from py2mint.footprints import attrs_used_by_method
