@@ -34,6 +34,14 @@ class C:
 ###############################################################################
 
 
+def times_2(x):
+    return x * 2
+
+
+def plus_1(z):
+    return z + 1
+
+
 def add(a, b: float = 0.0) -> float:
     """Adds numbers"""
     return a + b
@@ -43,11 +51,20 @@ def mult(x: float, y=1):
     return x * y
 
 
-# This one has every of the 4 combinations of (default y/n, annotated y/n)
 def formula1(w, /, x: float, y=1, *, z: int = 1):
+    """Has 3 kinds & every of the 4 combinations of (default y/n, annotated y/n)"""
     return ((w + x) * y) ** z
 
 
+def formula2(w, /, x: float, y=2, *, z: int = 3):
+    """Has 3 kinds & every of the 4 combinations of (default y/n, annotated y/n)
+
+    >>> formula2(0, 1, 2)
+    '(w:=0) + (x:=1) * (y:=2) ** (z:=3) == 8'
+    """
+    return f'(w:={w}) + (x:={x}) * (y:={y}) ** (z:={z}) == {w + x * y ** z}'
+
+
 # Only *args and **kwargs. Also has a doc with a doc test
 def sum_of_args(*args, **kwargs):
     """Sums the args, and returns a copy of kwargs with the sum of args added to each value.
