@@ -15,10 +15,69 @@ import pytest
 from functools import reduce
 from typing import Any
 
+from i2 import Sig
 from i2.signatures import *
-from i2.signatures import normalized_func
+from i2.signatures import (
+    normalized_func,
+    sigs_for_sigless_builtin_name,
+    _robust_signature_of_callable,
+)
 
-from i2.tests.util import sig_to_inputs, trace_call
+from i2.tests.util import (
+    call_and_return_error,
+    sig_to_inputs,
+    trace_call,
+    function_is_compatible_with_signature,
+)
+
+
+def test_signature_equality_and_hashing():
+
+    import pickle
+
+    def foo(x, y, z=3):
+        pass
+
+    def bar(x, y, z=3):
+        pass
+
+    # different instances of (should be a) same signature
+    ref_sig = Sig(foo)
+
+    sigs = [
+        ref_sig,
+        Sig(foo),  # another instance of the same Sig(foo)
+        Sig('(x, y, z=3)'),  # signature made explicitly from a string
+        Sig('x') + Sig('(y, z=3)'),  # signature made from an add operation
+        Sig(bar),  # different function, same signature
+        pickle.loads(pickle.dumps(ref_sig)),  # un-pickled signature
+    ]
+
+    assert all(
+        sig == ref_sig for sig in sigs
+    ), 'sigs should be equal from a == point of view'
+
+    # Let's see that
+    assert all(
+        hash(sig) == hash(ref_sig) for sig in sigs
+    ), 'sigs should have the same hash'
+
+    # ... and if that wasn't convincing, let's see how the sigs behave as dict keys:
+    # If sigs were different, the following would add key-value pairs to the base dict.
+    # But it doesn't. You always get the same one key (same signature) with different
+    # values:
+    t = {ref_sig: 0}
+    assert t == {ref_sig: 0}
+    t[sigs[2]] = 2
+    assert t == {ref_sig: 2}  # same signature with new 2 value
+    t[sigs[3]] = 3
+    assert t == {ref_sig: 3}  # same signature with new 3 value
+
+    # What if we just have a return annotation?
+    def baz(x, y, z=3) -> None:
+        pass
+
+    assert Sig(baz) != ref_sig, 'return annotation of baz should make it different'
 
 
 def test_signature_of_partial():
@@ -34,6 +93,19 @@ def test_signature_of_partial():
     assert str(Sig(partial(foo, 1, b=2))) == '(*, b=2, c=3) -> int'
 
 
+def test_some_edge_cases_of_sig():
+    from operator import itemgetter, attrgetter, methodcaller
+
+    assert Sig(itemgetter).names == ['key', 'keys']
+    assert Sig(itemgetter(1)).names == ['iterable']
+    assert Sig(itemgetter(1, 2)).names == ['iterable']
+    assert Sig(attrgetter).names == ['key', 'keys']
+    assert Sig(attrgetter('foo')).names == ['iterable']
+    assert Sig(attrgetter('foo', 'bar')).names == ['iterable']
+    assert Sig(methodcaller).names == ['name', 'args', 'kwargs']
+    # assert Sig(methodcaller('foo')).names == []  # fix!!
+
+
 def test_sig_wrap_edge_cases():
     """Tests some edge cases involving simultaneous changes of defaults and kinds.
 
@@ -62,6 +134,8 @@ def test_sig_wrap_edge_cases():
     def foo(x, y, z=0):
         return x + y * z
 
+    assert Sig(foo)._defaults_ == (3,)
+
     assert foo(1, 2) == 7
     #    works because Sig also changed __defaults__:
     assert foo.__defaults__ == (3,)
@@ -74,6 +148,8 @@ def test_sig_wrap_edge_cases():
     #    works because Sig also changed __defaults__ and __kwdefaults__:
     assert foo.__defaults__ == ()
     assert foo.__kwdefaults__ == {'z': 3}
+    assert Sig(foo)._defaults_ == ()
+    assert Sig(foo)._kwdefaults_ == {'z': 3}
 
     # The following (where we go from a (same kind) param not having a default,
     # to having one, also work
@@ -91,6 +167,8 @@ def test_sig_wrap_edge_cases():
 
     assert foo(1, 2) == 7
     assert foo(1, 2, z=10) == 21
+    assert Sig(foo)._defaults_ == ()
+    assert Sig(foo)._kwdefaults_ == {'z': 3}
 
 
 def test_tuple_the_args():
@@ -527,7 +605,7 @@ def test_call_forgivingly(sig_spec):
         # print(output)
         assert output == expected_output
 
-    for args, kwargs in sig_to_inputs(sig, ignore_variadics=True):
+    for args, kwargs in sig_to_inputs(sig, variadics_source=((), {})):
         kwargs = dict(kwargs, **{'some': 'extra', 'added': 'kwargs'})
         po_pk_count = sum(kind <= PK for kind in sig.kinds.values())
         if len(args) == po_pk_count:
@@ -588,7 +666,7 @@ def test_call_compatibility(sig_spec1, sig_spec2):
     #     pass
 
     pos1, pks1, vp1, kos1, vk1 = sig1.detail_names_by_kind()
-    for args, kwargs in sig_to_inputs(sig1, ignore_variadics=True):
+    for args, kwargs in sig_to_inputs(sig1, variadics_source=((), {})):
         if vp1 is not None and len(args) == len(pos1) + len(pks1):
             args += ('extra_arg',)
         if vk1 is not None:
@@ -602,3 +680,52 @@ def test_call_compatibility(sig_spec1, sig_spec2):
                 return
     if not is_compatible:
         raise AssertionError('sig1 is not compatible with sig2, when it should.')
+
+
+def test_bool():
+    name = 'bool'
+
+    sig = Sig(sigs_for_sigless_builtin_name[name])
+
+    assert function_is_compatible_with_signature(bool, sig)
+
+
+def visualize_errors_for_function_call(func, sig):
+    """
+    Calls func on sig_to_inputs and prints error if any
+    """
+
+    print(f'============================================================{str(func)}')
+    for args, kwargs in sig_to_inputs(sig):
+        result = call_and_return_error(func, *args, **kwargs)
+        if result is not None:
+            print(f'{args=} , {kwargs=}')
+            print(result)
+
+
+def test_sigless_builtins():
+    from operator import itemgetter, attrgetter, methodcaller
+
+    mapping_methods = {
+        '__eq__',
+        '__ne__',
+        '__iter__',
+        '__getitem__',
+        '__len__',
+        '__contains__',
+        '__setitem__',
+        '__delitem__',
+    }
+    special_cases = {'breakpoint'} | mapping_methods
+
+    for name in sigs_for_sigless_builtin_name:
+        # removed breakpoint as it triggers a pdb session
+        if name in special_cases:
+            continue
+        sig = Sig(sigs_for_sigless_builtin_name[name])
+        assert function_is_compatible_with_signature(eval(name), sig)
+
+    d = {'a': 1, 'b': 2}
+    for name in mapping_methods:
+        method = getattr(d, name)
+        assert function_is_compatible_with_signature(method, Sig(method))
