@@ -1,9 +1,21 @@
 """Utils for testing"""
 
 from inspect import Parameter, Signature
-from typing import List, Any, Union, Callable, Iterator, Tuple, Optional, Iterable
+from itertools import product
+from typing import (
+    List,
+    Any,
+    Union,
+    Callable,
+    Iterator,
+    Tuple,
+    Optional,
+    Iterable,
+    Mapping,
+)
+from inspect import signature
 
-from i2.signatures import KO, PK, PO, var_param_kinds
+from i2.signatures import KO, PK, PO, VP, VK, var_param_kinds
 from i2.signatures import _empty
 from i2.signatures import ParamsAble, Sig, ensure_param, SignatureAble
 
@@ -11,6 +23,15 @@ ParameterAble = Union[int, Parameter, str]
 ParamsAble_ = Union[ParamsAble, str, List[int]]
 
 
+def _is_valid_arg_for_sig(x):
+    return (
+        isinstance(x, (Callable, Signature))
+        or isinstance(x, str)
+        and x.startswith('(')
+        and x.endswith(')')
+    )
+
+
 def generate_params(params: ParamsAble_):
     """Generate inspect.Parameter instances quickly.
 
@@ -22,7 +43,7 @@ def generate_params(params: ParamsAble_):
     >>> str(Sig(generate_params("00111234")))
     '(a00, a01, /, a12, a13, a14, *a25, a36, **a47)'
     """
-    if isinstance(params, (Callable, Signature)):
+    if _is_valid_arg_for_sig(params):
         # generate params from a callable or signature
         yield from Sig(params).params
     else:
@@ -95,6 +116,10 @@ def _str_of_call_args(_call_kwargs: dict):
     return ', '.join(f'{k}={v}' for k, v in _call_kwargs.items())
 
 
+def _params_to_name(params):
+    return 'f' + ''.join(str(int(p.kind)) for p in params)
+
+
 def mk_func_from_params(
     params: ParamsAble = '00111234',
     *,
@@ -104,6 +129,8 @@ def mk_func_from_params(
 ):
     """Make a function (that actually returns something based on args) from params.
 
+    See Also: ``sig_to_func``
+
     :param params: params (arguments) of the function (can be expressed in many ways!)
     :param defaults: Optional {argname: default,...} dict to inject defaults
     :param name: Optional name to give the function
@@ -187,19 +214,206 @@ def mk_func_from_params(
         )
         return callback(_call_kwargs)
 
-    arg_str_func.__name__ = name or 'f' + ''.join(str(p.kind) for p in params)
+    arg_str_func.__name__ = name or _params_to_name(params)
 
     return arg_str_func
 
 
+def _sig_to_str_of_call_args_code_str(sig: Sig):
+    return (
+        'return ' + 'f"' + _str_of_call_args({p: f'{{{p}}}' for p in sig.names}) + '"'
+    )
+
+
+def _is_simple_expression(code_lines):
+    if len(code_lines) == 1:
+        line = code_lines[0].strip()
+        if not (
+            line in {'pass', '...'}
+            or line.startswith('return')
+            or (line.startswith('"') and line.endswith("'"))
+        ):
+            return True
+    else:
+        return False
+
+
+def sig_to_func(
+    sig: ParamsAble = '00111234',
+    code_lines: Union[
+        str, Iterable, Callable[[Sig], str]
+    ] = _sig_to_str_of_call_args_code_str,
+    *,
+    name: Optional[str] = None,
+    globals: Optional[dict] = None,
+    locals: Optional[Mapping] = None,
+):
+    """
+    Make a function from a signature
+
+    See Also: ``mk_func_from_params``
+
+    More information: https://github.com/i2mint/i2/issues/34
+
+    :param sig: Signature (or something that can be made into a signature) the func
+    should have
+    :param name: Name the function should have
+    :param code_lines: The code lines
+
+    The ``globals`` must be a dictionary and ``locals`` can be any mapping, defaulting
+    to the current globals and locals. If only ``globals`` is given, ``locals``
+    defaults to it.
+
+    >>> sig = Sig('(a, /, b=2)')
+    >>> f = sig_to_func(sig)
+    >>> str(Sig(f))
+    '(a, /, b=2)'
+    >>> f(2, 3)
+    'a=2, b=3'
+    >>> f.__name__
+    'f01'
+
+    See how the function was given a name automatically? This name was created by
+    appending the kind number
+    (see https://docs.python.org/3/library/inspect.html#inspect.Parameter.kind)
+    to 'f'.
+
+    Let's demo how we can
+
+    - use kind numbers (see `inspect` module) to specify the signature
+
+    - give an explicit name to the function
+
+    - specify ``locals()`` so that ``sig_to_func`` can insert the new function there
+    (which makes it picklable, for instances)
+
+    >>> _ = sig_to_func('012', name='foo', locals=locals())
+    >>> # and now `foo` is in local name space, and has signature:
+    >>> str(Sig(foo))
+    '(a00, /, a11, *a22)'
+    >>> foo(1,2,3,4,5)
+    'a00=1, a11=2, a22=(3, 4, 5)'
+
+    In the examples above, the function body was was created from the input ``sig``
+    through the ``_sig_to_str_of_call_args_code_str`` function which outputs a string
+    formed from the input argument names and values.
+
+    But you can specify your own function. This function should take a ``Sig`` object
+    and return the string or lines (iterable of strings) of the function's body.
+    You can also specify a string or lines directly.
+
+    >>> g = sig_to_func('(x, y=2)', 'pass')
+    >>> assert g(2, y=3) is None
+    >>> h = sig_to_func('(x, y=2)', 'return x * y')
+    >>> h(3)
+    6
+    >>> h(x=10, y=3)
+    30
+
+    You can even omit the return instruction if, as in ``lambda`` functions, the body is
+    a simple expression (not `pass`, `...`, or starting something with quotes).
+
+    >>> h = sig_to_func('(x, y=2)', 'x / y')
+    >>> h(10, 5)
+    2.0
+
+    """
+    sig = Sig(generate_params(sig))
+    name = name or _params_to_name(sig.params)
+    if callable(code_lines):
+        code_lines = code_lines(sig)  # call the function on sig to get lines
+    if isinstance(code_lines, str):
+        code_lines = code_lines.split('\n')
+    if _is_simple_expression(code_lines):
+        # If code_lines has only one line and it seems it's an expression, prepend return
+        code_lines = [f'return {code_lines[0]}']
+    code_string = '\n\t'.join(code_lines)
+    func_def_string = f'def {name}{sig}:\n\t{code_string}'
+    _locals = locals or {}
+    exec(func_def_string, globals, _locals)
+    return _locals[name]
+
+
+def _all_prefixes(x: Iterable):
+    """
+    >>> list(_all_prefixes([1,2,3]))
+    [(), (1,), (1, 2), (1, 2, 3)]
+    >>> list(map(dict, _all_prefixes({'a': 1, 'b': 2}.items())))
+    [{}, {'a': 1}, {'a': 1, 'b': 2}]
+    """
+    x = tuple(x)
+    for i in range(len(x) + 1):
+        yield x[:i]
+
+
+def _args_kwargs_combinations(args, kwargs):
+    """
+
+    >>> assert list(_args_kwargs_combinations((1,2), {'a': 3, 'b': 4})) == [
+    ...     ((), {}),
+    ...     ((), {'a': 3}),
+    ...     ((), {'a': 3, 'b': 4}),
+    ...     ((1,), {}),
+    ...     ((1,), {'a': 3}),
+    ...     ((1,), {'a': 3, 'b': 4}),
+    ...     ((1, 2), {}),
+    ...     ((1, 2), {'a': 3}),
+    ...     ((1, 2), {'a': 3, 'b': 4})
+    ... ]
+
+    """
+    for args_prefix, kwargs_items_prefix in product(
+        _all_prefixes(args), _all_prefixes(kwargs.items())
+    ):
+        yield args_prefix, dict(kwargs_items_prefix)
+
+
+def variadic_type(sig, variadics):
+    var_kinds = [sig.kinds[param] for param in variadics]
+    if not variadics:
+        return 'no_var'
+    if VP in var_kinds and VK not in var_kinds:
+        return 'vp_only'
+    if VK in var_kinds and VP not in var_kinds:
+        return 'vk_only'
+    else:
+        return 'vp_vk'
+
+
+def create_variadic_source(sig, variadics, dflt_source):
+    var_type = variadic_type(sig, variadics)
+    if var_type == 'no_var':
+        result = ((), {})
+    elif var_type == 'vp_only':
+        result = (dflt_source[0], {})
+    elif var_type == 'vk_only':
+        result = ((), dflt_source[1])
+    elif var_type == 'vp_vk':
+        result = dflt_source
+    return result
+
+
+# TODO: Get rid of ignore_variadics once using code is refactored
 def sig_to_inputs(
     sig: SignatureAble,
     argument_vals: Optional[Iterable] = None,
-    ignore_variadics: bool = False,
+    *,
+    variadics_source: Tuple[tuple, dict] = (
+        ('args1', 'args2'),
+        {'kwargs1': 'kwargs1_val'},
+    ),
 ) -> Iterator[Tuple[tuple, dict]]:
     """Generate all kind-valid (arg, kwargs) input combinations for a function with a
     given signature ``sig``, with argument values taken from the ``argument_vals``
 
+    :param sig: A signature or anything that ``i2.Sig`` can use to create one (e.g.
+        function, string, list of dicts etc.)
+    :param argument_vals: An interable from which the argument values will be drawn.
+        Defaults to ``list(range(n_args))``.
+    :param variadics_source: The ``(tuple, dict)`` pair that will be used to extract
+        variadics' inputs.
+    :return: A generator of ``(args: tuple, kwargs: dict)`` pairs
+
     >>> assert list(
     ...     sig_to_inputs(lambda a, b, /, c, d, *, e, f: None)
     ... ) == [
@@ -208,51 +422,49 @@ def sig_to_inputs(
     ...     ((0, 1, 2, 3), {'e': 4, 'f': 5})
     ... ]
 
-    >>> assert list(
-    ...     sig_to_inputs(
-    ...         Sig('(a=-1, /, b=-1, *args, c=-1, **kwargs)'),
-    ...         ignore_variadics=True
-    ...     )
-    ... ) == [
-    ...     ((), {}),
-    ...     ((0,), {}),
-    ...     ((), {'b': 0}),
-    ...     ((0,), {'b': 1}),
-    ...     ((0, 1), {}),
-    ...     ((), {'c': 0}),
-    ...     ((0,), {'c': 1}),
-    ...     ((), {'b': 0, 'c': 1}),
-    ...     ((0,), {'b': 1, 'c': 2}),
-    ...     ((0, 1), {'c': 2})
-    ... ]
+    >>> list(sig_to_inputs(Sig('(a, *args, b, **kwargs)')))  # doctest: +NORMALIZE_WHITESPACE
+    [((), {'a': 0, 'b': 1}),
+     ((), {'a': 0, 'b': 1, 'kwargs1': 'kwargs1_val'}),
+     (('args1',), {'a': 0, 'b': 1}),
+     (('args1',), {'a': 0, 'b': 1, 'kwargs1': 'kwargs1_val'}),
+     (('args1', 'args2'), {'a': 0, 'b': 1}),
+     (('args1', 'args2'), {'a': 0, 'b': 1, 'kwargs1': 'kwargs1_val'}),
+     ((0,), {'b': 1}),
+     ((0,), {'b': 1, 'kwargs1': 'kwargs1_val'}),
+     ((0, 'args1'), {'b': 1}),
+     ((0, 'args1'), {'b': 1, 'kwargs1': 'kwargs1_val'}),
+     ((0, 'args1', 'args2'), {'b': 1}),
+     ((0, 'args1', 'args2'), {'b': 1, 'kwargs1': 'kwargs1_val'})]
 
-    :param sig: A signature or anything that ``i2.Sig`` can use to create one (e.g.
-        function, string, list of dicts etc.)
-    :param argument_vals: An interable from which the argument values will be drawn.
-        Defaults to ``list(range(n_args))``.
-    :return: A generator of ``(args: tuple, kwargs: dict)`` pairs
+    Tip: To ignore variadics all together, you can specify ``variadics_source=((), {})``.
     """
     sig = Sig(sig)
+    init_sig = sig
     already_yielded = []
     variadics = [param for param, kind in sig.kinds.items() if kind in var_param_kinds]
     if variadics:
-        if ignore_variadics:
-            for v in variadics:
-                sig -= v
-        else:
-            raise ValueError(f'Not allowed to have variadics: {sig}')
-    for sub_sig in _get_sub_sigs_from_default_values(sig):
-        po, pk, ko = _get_non_variadic_kind_counts(sub_sig)
-        for args, kwargs_vals in _sig_to_inputs(
-            po, pk, ko, argument_vals=argument_vals
-        ):
-            input_ = (
-                tuple(args),
-                {k: v for k, v in zip(sub_sig.names[len(args) :], kwargs_vals)},
-            )
-            if input_ not in already_yielded:
-                yield input_
-                already_yielded.append(input_)
+        for v in variadics:
+            sig -= v
+        variadics_source = create_variadic_source(
+            init_sig, variadics, dflt_source=variadics_source
+        )
+        var_args, var_kwargs = variadics_source
+        for args, kwargs in sig_to_inputs(sig, argument_vals):
+            for _args, _kwargs in _args_kwargs_combinations(var_args, var_kwargs):
+                yield args + _args, dict(kwargs, **_kwargs)
+    else:
+        for sub_sig in _get_sub_sigs_from_default_values(sig):
+            po, pk, ko = _get_non_variadic_kind_counts(sub_sig)
+            for args, kwargs_vals in _sig_to_inputs(
+                po, pk, ko, argument_vals=argument_vals
+            ):
+                input_ = (
+                    tuple(args),
+                    {k: v for k, v in zip(sub_sig.names[len(args) :], kwargs_vals)},
+                )
+                if input_ not in already_yielded:
+                    yield input_
+                    already_yielded.append(input_)
 
 
 def _sig_to_inputs(po=0, pk=0, ko=0, argument_vals: Optional[Iterable] = None):
@@ -280,7 +492,7 @@ def _sig_to_inputs(po=0, pk=0, ko=0, argument_vals: Optional[Iterable] = None):
 
 def _get_sub_sigs_from_default_values(sig: Sig) -> Iterator[Sig]:
     """Generate all the signatures compatible with the given signature
-    by ignoring the arguments that have default values. 
+    by ignoring the arguments that have default values.
     >>> sig = Sig('(a=0, /, b=0, *args, c=0, **kwargs)')
     >>> assert [str(s) for s in _get_sub_sigs_from_default_values(sig)] == [
     ...     '(*args, **kwargs)',
@@ -341,6 +553,94 @@ def mk_func_inputs_for_params(params: ParamsAble_, param_to_input):
     pass
 
 
+# ---------------------------------------------------------------------------------------
+# Tools to analyze compatibility between signature and function call
+
+from i2 import Pipe
+import re
+
+_signature_msg_patterns = [
+    'keyword arguments$',
+    'invalid keyword argument',
+    'expected at most',
+    'keyword argument',
+    'got some positional\-only arguments passed as keyword arguments',
+    'no signature found',
+]
+
+_signature_msg_regex = re.compile('|'.join(map('({})'.format, _signature_msg_patterns)))
+is_signature_msg = Pipe(_signature_msg_regex.search, bool)
+
+
+def _is_signature_error(
+    error_obj,
+    signature_error_types=(TypeError, ValueError),
+    is_signature_msg=is_signature_msg,
+):
+    if isinstance(error_obj, signature_error_types):
+        error_msg = str(error_obj)
+        return is_signature_msg(error_msg)
+    return False
+
+
+def call_and_return_error(func, /, *args, **kwargs):
+    try:
+        func(*args, **kwargs)
+        return None
+    except BaseException as error_obj:
+        return error_obj
+
+
+def on_error_return_none(func, /, *args, **kwargs):
+    try:
+        return func(*args, **kwargs)
+    except BaseException as error_obj:
+        return None
+
+
+call_raises_signature_error = Pipe(call_and_return_error, _is_signature_error)
+
+call_raises_signature_error.__doc__ = '''
+>>> call_raises_signature_error(lambda x, /, y: None, 1, y=2)
+False
+>>> call_raises_signature_error(lambda x, /, y: None, x=1, y=2)
+True
+'''
+
+# Yes, I too see that this can be made into yet another Pipe!
+def function_is_compatible_with_signature(func, sig):
+    """
+    Runs through all combinations of positional and keyword arguments,
+
+    >>> function_is_compatible_with_signature(hasattr, Sig(lambda obj, name: ...))
+    False
+    >>> function_is_compatible_with_signature(hasattr, Sig(lambda obj, name, /: ...))
+    True
+    """
+
+    def _call_raises_sig_error():
+        for args, kwargs in sig_to_inputs(sig):
+            yield call_raises_signature_error(func, *args, **kwargs)
+
+    return not any(_call_raises_sig_error())
+
+
+def builtin_objects():
+    for name in dir(__builtins__):
+        yield getattr(__builtins__, name)
+
+
+def builtin_signatureless_callables():
+    """
+    A generator of builtin callables that don't have signatures.
+    """
+    for obj in builtin_objects():
+        if callable(obj) and call_raises_signature_error(signature, obj):
+            yield obj
+
+
+# ---------------------------------------------------------------------------------------
+
 #
 # @mk_func_from_params.register
 # def mk_func_from_params(params: Iterable[int], defaults=None, name=None):
