@@ -2,9 +2,10 @@
 from operator import attrgetter, itemgetter
 import inspect
 import re
+import os
 import itertools
 import sys
-import functools
+from functools import partial
 import types
 from typing import (
     Mapping,
@@ -16,14 +17,159 @@ from typing import (
     Iterable,
     TypeVar,
 )
+import contextlib
+import io
 
 T = TypeVar('T')
 
 
+@contextlib.contextmanager
+def FileLikeObject(file, *, io_cls=io.BytesIO, open_mode='rb'):
+    """Context manager for file-like objects.
+
+    The purpose of this context manager is to be able to ensure we have a file-like 
+    object interface to work with, regardless of whether we are given a file path,
+    bytes of a file, or an open file pointer.
+
+    Args:
+        file (str, bytes, io.IOBase): The file path, bytes of a file, or an open file pointer.
+
+    Yields:
+        io.IOBase: A file-like object.
+    """
+    if isinstance(file, str):
+        # If file is a string, open the file and yield the file pointer
+        with open(file, open_mode) as f:
+            yield f  # TODO: Is f protected by the open context?
+    elif isinstance(file, bytes):
+        # If file is bytes, create a BytesIO object and yield it
+        with io.BytesIO(file) as f:
+            yield f
+    else:
+        # Otherwise, assume file is already a file-like object and yield it
+        yield file
+
+
+def copy_func(
+    func: Callable, *, copy_dict: bool = True, code=None, globals_: dict = None
+):
+    """Make a (shallow) copy of a function.
+
+    >>> f = lambda x, *, y=2: x * y
+    >>> f.an_attr = 42
+    >>> f_copy = copy_func(f)
+    >>> f_copy(3) == f(3) == 6
+    True
+    >>> f_copy.an_attr == f.an_attr == 42
+    True
+
+    Verify that making an attribute in one won't create an attribute in the other:
+
+    >>> f.another_attr = 42
+    >>> hasattr(f_copy, 'another_attr')
+    False
+    >>> f_copy.yet_another_attr = 84
+    >>> hasattr(f, 'yet_another_attr')
+    False
+
+    :param func: The function to be copied.
+    :param copy_dict: Indicates whether to copy the ``__dict__`` attribute of the
+        function. Defaults to ``True``.
+    :param code: The value to be use as the ``__code__`` attribute of the copy.
+    :param globals_: The value to be use as the ``__globals__`` attribute of the copy.
+    :return: A shallow copy of the function.
+
+    Note that it should always work with proper functions and attempts to do the
+    best job it can with other callables, but there are no guarantees on how
+    ``copy_function`` will behave with custom callables.
+
+    If these custom callables don't have a `__code__` attribute, the copy will fail.
+    Furthermore, if the custom callable  doesn't have ``__globals__``, the empty
+    dictionary will be used as the globals.
+    We provide a ``code`` and ``globals`` argument to allow the user to provide
+    the ``__code__`` and ``__globals__`` attributes of the function to be copied.
+
+    :param func: The function to be copied. Must be a function, not just any method or callable.
+    :param copy_dict: Also copy any attributes set on the function instance. Defaults to ``True``.
+    :return: A shallow copy of the function.
+    """
+    from types import FunctionType
+
+    code = code or getattr(func, '__code__', None)
+    if not isinstance(code, types.CodeType):
+        raise TypeError(f'Expected a types.CodeType object, but got {type(code)=}')
+    globals_ = globals_ or getattr(func, '__globals__', {})
+    new_func = FunctionType(
+        code,  # if your func doesn't have a __code__ attr, can't make a copy!
+        globals_,
+        name=getattr(func, '__name__', None),
+        argdefs=getattr(func, '__defaults__', None),
+        closure=getattr(func, '__closure__', None),
+    )
+    if hasattr(func, '__kwdefaults__'):
+        new_func.__kwdefaults__ = func.__kwdefaults__
+    if copy_dict:
+        new_func.__dict__.update(func.__dict__)
+    return new_func
+
+
 class OverwritesForbidden(ValueError):
     """Raise when a user is not allowed to overwrite a mapping's key"""
 
 
+def is_lambda(func):
+    return getattr(func, '__name__', None) == '<lambda>'
+
+
+# TODO: Fragile. Make more robust.
+def lambda_code(lambda_func) -> str:
+    """Extract code of expression from lambda function.
+    For lambda code-extraction see:
+    https://stackoverflow.com/questions/73980648/how-to-transform-a-lambda-function-into-a-pickle-able-function
+
+    """
+    func_str = str(inspect.getsourcelines(lambda_func)[0])
+    return func_str.strip("['\\n']").split(' = ')[1]
+
+
+# TODO: Only works with lambdas so either assert function is a lambda on construction
+#  or make it work with functions more generally.
+class PicklableLambda:
+    """
+    Wraps a lambda function to make it picklable (through extracting its code)
+    Also, provide it with a name, optionally.
+
+    >>> f = lambda x, y=0: x + y
+    >>> ff = PicklableLambda(f)
+    >>> import pickle
+    >>> fff = pickle.loads(pickle.dumps(ff))
+    >>> assert fff(2, 3) == ff(2, 3) == f(2, 3)
+
+    For lambda code-extraction see:
+    https://stackoverflow.com/questions/73980648/how-to-transform-a-lambda-function-into-a-pickle-able-function
+
+    """
+
+    def __init__(self, func, name=None):
+        self.func = func
+        self.__signature__ = inspect.signature(self.func)
+        self.__name__ = name or getattr(func, '__name__', type(self).__name__)
+
+    def __getstate__(self):
+        return lambda_code(self.func), self.__name__
+
+    def __setstate__(self, state):
+        func_code, name = state
+        func = eval(func_code)  # scary?
+        self.__init__(func, name)
+
+    def __call__(self, *args, **kwargs):
+        return self.func(*args, **kwargs)
+
+    def __repr__(self):
+        return f'<{type(self).__name__}({self.__name__})>'
+
+
 def ensure_identifiers(
     *objs: Iterable[T],
     get_identfiers: Callable[[T], Iterable[str]] = str.split,
@@ -135,10 +281,10 @@ def insert_name_based_objects_in_scope(
             )
 
 
-class Literal:
+class LiteralVal:
     """An object to indicate that the value should be considered literally.
 
-    >>> t = Literal(42)
+    >>> t = LiteralVal(42)
     >>> t.get_val()
     42
     >>> t()
@@ -236,7 +382,7 @@ def path_extractor(tree, path, getter=dflt_idx_preprocessor, *, path_sep='.'):
         idx, *path = path  # extract path[0] as idx & update path to path[1:]
         if isinstance(idx, str) and idx == '*':
             idx = lambda x: True  # use a filter function (but filter everything in)
-        if callable(idx) and not isinstance(idx, Literal):
+        if callable(idx) and not isinstance(idx, LiteralVal):
             # If idx is a non-literal callable, consider it as a filter to be applied
             # to iter(tree)
             # TODO: https://github.com/i2mint/i2/issues/27
@@ -244,7 +390,7 @@ def path_extractor(tree, path, getter=dflt_idx_preprocessor, *, path_sep='.'):
                 path_extractor(sub_tree, path, getter) for sub_tree in filter(idx, tree)
             ]
         else:
-            if isinstance(idx, Literal):
+            if isinstance(idx, LiteralVal):
                 # Use of Literal is meant get out of trouble if we want to use a
                 # callable as an actual index, not as a filter.
                 idx = idx.get_val()
@@ -265,6 +411,32 @@ def dp_get(d, dot_path):
     return path_extractor(d, dot_path, lambda d, k: d[k])
 
 
+def get_app_data_folder():
+    """
+    Returns the full path of a directory suitable for storing application-specific data.
+
+    On Windows, this is typically %APPDATA%.
+    On macOS, this is typically ~/.config.
+    On Linux, this is typically ~/.config.
+
+    Returns:
+        str: The full path of the app data folder.
+
+    See https://github.com/i2mint/i2mint/issues/1.
+    """
+    if os.name == 'nt':
+        # Windows
+        app_data_folder = os.getenv('APPDATA')
+    elif os.name == 'darwin':
+        # macOS
+        app_data_folder = os.path.expanduser('~/.config')
+    else:
+        # Linux/Unix
+        app_data_folder = os.path.expanduser('~/.config')
+
+    return app_data_folder
+
+
 class lazyprop:
     """
     A descriptor implementation of lazyprop (cached property) from David Beazley's "Python Cookbook" book.
@@ -634,6 +806,8 @@ def _indent(text, margin, newline='\n', key=bool):
 
 NO_DEFAULT = mk_sentinel('NO_DEFAULT', boolean_value=False)
 
+# --------------------------------------------------------------------------------------
+# FunctionBuilder, vendored and adapted from boltons.funcutils
 
 from inspect import formatannotation
 
@@ -860,7 +1034,7 @@ class FunctionBuilder(object):
         if not callable(func):
             raise TypeError('expected callable object, not %r' % (func,))
 
-        if isinstance(func, functools.partial):
+        if isinstance(func, partial):
             kwargs = {
                 'name': func.__name__,
                 'doc': func.__doc__,
@@ -1006,7 +1180,6 @@ class FunctionBuilder(object):
         return
 
     def _compile(self, src, execdict):
-
         filename = '<%s-%d>' % (self.filename, next(self._compile_count),)
         try:
             code = compile(src, filename, 'single')
