@@ -68,6 +68,7 @@ from inspect import Parameter, signature
 from typing import (
     Mapping,
     Callable,
+    Any,
     Optional,
     Union,
     Iterable,
@@ -78,8 +79,20 @@ from typing import (
 )
 
 from types import MethodType
+from inspect import Parameter
+from dataclasses import make_dataclass, dataclass
 
-from i2.signatures import Sig, name_of_obj, KO, PK, VK
+from i2.signatures import (
+    Sig,
+    name_of_obj,
+    KO,
+    PK,
+    VK,
+    parameter_to_dict,
+    call_forgivingly,
+    _call_forgivingly,
+)
+from i2.errors import OverwritesNotAllowed
 from i2.multi_object import Pipe
 from i2.deco import double_up_as_factory
 
@@ -125,7 +138,6 @@ class MakeFromFunc:
         return self.func_to_obj(func)
 
 
-# from wrapt import decorator
 # TODO: Continue factoring out Wrap and Wrapx code
 class _Wrap:
     """To be used as the base of actual Wrap objects."""
@@ -140,7 +152,7 @@ class _Wrap:
         # have had. It should be before anything else so it doesn't overwrite stuff
         # that we may add to self in init (like .func for example!)
         wraps(func)(self)  # TODO: should we really copy everything by default?
-        if name := kwargs.get('name', None) is not None:
+        if (name := kwargs.get('name', None)) is not None:
             self.__name__ = name
         self.func = func  # Note: overwrites self.func that wraps MAY have inserted
         self.__wrapped__ = func
@@ -178,11 +190,22 @@ class _Wrap:
 
     # To get help(instance.method) to work!
     # TODO: Does this have undesirable side effects?
+    # TODO: Perhaps self.__call__.__code__ is better (since it's actually what runs)?
     @property
     def __code__(self):
         return self.func.__code__
 
 
+def _defaults_and_kwdefaults_of_func(func: Callable):
+    try:
+        return func.__defaults__, func.__kwdefaults__
+    except AttributeError:
+        # if python can't do it on it's own, you may have a special kind of callable
+        # here, so try doing it through Sig, which is more flexible
+        sig = Sig(func)
+        return sig._defaults_, sig._kwdefaults_
+
+
 class Wrap(_Wrap):
     """A function wrapper with interface modifiers.
 
@@ -249,7 +272,6 @@ class Wrap(_Wrap):
     >>> wrapped_func(2, 'Hi', 'world!')
     'Hiworld!Hiworld! ITSME!!!'
 
-
     A more involved example:
 
     >>> def ingress(a, b: str, c="hi"):
@@ -271,13 +293,11 @@ class Wrap(_Wrap):
     ...     == len("Hi world! Hi world! Hi world! ")
     ... )
 
-    An ingress function links the interface of the wrapper to the interface of the
+    An ``ingress`` function links the interface of the wrapper to the interface of the
     wrapped func; therefore it's definition often depends on information of both,
     and for that reason, we provide the ability to specify the ingress not only
     explicitly (as in the examples above), but through a factory -- a function that
-    will be called on `func` to produce the ingress that should be used to wrap it.
-
-
+    will be called on ``func`` to produce the ingress that should be used to wrap it.
 
     .. seealso::
 
@@ -291,6 +311,9 @@ class Wrap(_Wrap):
 
         if ingress is None:
             self.ingress = transparent_ingress
+            self.__defaults__, self.__kwdefaults__ = _defaults_and_kwdefaults_of_func(
+                func
+            )
         else:
 
             if isinstance(ingress, MakeFromFunc):
@@ -298,11 +321,14 @@ class Wrap(_Wrap):
                 func_to_ingress = ingress  # it's not the ingress function itself
                 # ... but an ingress factory: Should make the ingress in function of func
                 self.ingress = func_to_ingress(func)
-
             else:
                 assert callable(ingress), f'Should be callable: {ingress}'
                 self.ingress = ingress
+
             ingress_sig = Sig(self.ingress)
+            self.__defaults__, self.__kwdefaults__ = _defaults_and_kwdefaults_of_func(
+                self.ingress
+            )
 
         return_annotation = empty
 
@@ -457,24 +483,20 @@ def append_empty_args(func):
 class Ingress:
     """The Ingress class offers a template for creating ingress classes.
 
-    Note that when writing a decorator with i2.wrapper, you're usually better off
+    Note that when writing a decorator with ``i2.wrapper``, you're usually better off
     writing an ingress function for the purpose. As a result, your code will usually
     be less complex, easier to read, and more efficient than using the Ingress class.
 
-    So why use the Ingress class at all? For one, because it'll take care of some common
-    mechanics for you, so once you understand how to use it,
-    you'll probably create a correct wrapper faster.
+    So why use the ``Ingress`` class at all? For one, because it'll take care of some
+    common mechanics for you, so once you understand how to use it, you'll probably
+    create a correct wrapper faster.
 
     Further, if you're writing a general wrapping tool (e.g. your own currying machine,
     some rule-based input casting function, etc.) then you'll find that using
     Ingres will usually with on the complexity, readability and/or efficiency front.
 
-    >>> from i2.wrapper import Ingress, wrap
-    >>> from inspect import signature
-    >>> from i2.wrapper import InnerMapIngress
-    >>>
-    >>> # Ingress = InnerMapIngress
-    >>>
+    Consider the following function:
+
     >>> def f(w, /, x: float, y=2, *, z: int = 3):
     ...     return f"(w:={w}) + (x:={x}) * (y:={y}) ** (z:={z}) == {w + x * y ** z}"
     >>>
@@ -487,6 +509,9 @@ class Ingress:
     in that case.
     Say you wanted all input values to be cast to ints. In that case you could do:
 
+    >>> from i2.wrapper import Ingress, wrap
+    >>> from inspect import signature
+    >>>
     >>> trans_all_vals_to_ints = lambda d: {k: int(v) for k, v in d.items()}
     >>>
     >>> cli_f = wrap(
@@ -524,11 +549,12 @@ class Ingress:
     above.
     Instead, they can be
 
-    - ``1-to-many`` (e.g. the outer 'w' is used to compute the inner ``w`` and ``x``)
+    - ``1-to-many`` (e.g. the outer ``w`` is used to compute the inner ``w`` and ``x``)
 
-    - ``many-to-1 (e.g. the outer ``x`` and ``y`` are used to compute inner ``y``)
+    - ``many-to-1`` (e.g. the outer ``x`` and ``y`` are used to compute inner ``y``)
 
     .. code-block::
+
           w   x   y   z
          / \   \ /    |
         w   x   y     z
@@ -583,7 +609,14 @@ class Ingress:
     """
 
     def __init__(
-        self, inner_sig, kwargs_trans: Optional[KwargsTrans] = None, outer_sig=None
+        self,
+        inner_sig,
+        kwargs_trans: Optional[KwargsTrans] = None,
+        outer_sig=None,
+        *,
+        allow_excess=True,
+        apply_defaults=True,
+        allow_partial=False,
     ):
         """Init of an Ingress instance.
 
@@ -595,14 +628,20 @@ class Ingress:
             Note that though both outer and inner signatures could have those annoying
             position-only kinds, you don't have to think of that.
             The parameter kind restrictions are taken care of automatically.
-        :param outer_sig: The outer signature. The signature the ingress function
-            (there for the wrapped function) will have. Also serves to convert input
-            (args, kwargs) to the kwargs that will be given to kwargs_trans.
+        :param outer_sig: The outer signature. The ingress function's signature will
+            have (and therefore, the wrapped function's signature too.
+            Also serves to convert input ``(args, kwargs)`` to the ``kwargs``
+            to the kwargs that will be given to ``kwargs_trans``.
+        :param allow_excess: Whether to allow the inner kwargs to have some excess
+            variables in them. This enables more flexibility in the outer signature,
+            but may want to set to ``False`` to be more explicit.
+        :param apply_defaults: Whether to apply the defaults of the outer signature.
+            Default is ``True`` but in some rare cases, you may not want to apply them.
 
         When making an Ingress function directly, one must take care that
         ``inner_sig``, ``kwargs_trans`` and ``outer_sig`` are aligned.
 
-        Namely, 'kwargs_trans' must be able to handle outputs of
+        Namely, ``kwargs_trans`` must be able to handle outputs of
         ``outer_sig.kwargs_from_args_and_kwargs`` and itself output kwargs that
         can be handled by ``inner_sig.args_and_kwargs_from_kwargs``.
 
@@ -621,11 +660,14 @@ class Ingress:
         self.outer_sig = Sig(outer_sig)
 
         self.outer_sig(self)
+        self.apply_defaults = apply_defaults
+        self.allow_excess = allow_excess
+        self.allow_partial = allow_partial
 
     def __call__(self, *ingress_args, **ingress_kwargs):
         # Get the all-keywords version of the arguments (args,kwargs->kwargs)
         func_kwargs = self.outer_sig.kwargs_from_args_and_kwargs(
-            ingress_args, ingress_kwargs, apply_defaults=True
+            ingress_args, ingress_kwargs, apply_defaults=self.apply_defaults
         )
 
         func_kwargs = dict(
@@ -635,11 +677,12 @@ class Ingress:
 
         # Return an (args, kwargs) pair that respects the inner function's
         # argument kind restrictions.
-        # Note: Originally was with (default) allow_excess=False. Changed to True to
-        #       allow more flexibility in outer sig. But is this sane? Worth it?
         # TODO: Reflect on pros/cons of allow_excess=True
         return self.inner_sig.args_and_kwargs_from_kwargs(
-            func_kwargs, apply_defaults=True, allow_excess=True
+            func_kwargs,
+            apply_defaults=True,
+            allow_excess=self.allow_excess,
+            allow_partial=self.allow_partial,
         )
 
     def __repr__(self):
@@ -718,9 +761,6 @@ def invert_map(d: dict):
         raise ValueError(f'There are duplicate keys so I can invert map: {d}')
 
 
-from i2.signatures import parameter_to_dict
-
-
 def parameters_to_dict(parameters):
     return {name: parameter_to_dict(param) for name, param in parameters.items()}
 
@@ -791,7 +831,7 @@ class InnerMapIngress:
     - where ``y`` was named ``you`` instead, and has an annotation (``int``).
 
     - where the default of ``z`` was ``10`` instead of ``3``, and doesn't have an
-    annotation.
+        annotation.
 
     In order to get a version of this function we wanted (more lenient kinds,
     with some annotations and a default change), we can use the ingress function:
@@ -903,18 +943,25 @@ class InnerMapIngress:
             **self.kwargs_trans(func_kwargs),  # change those that kwargs_trans desires
         )
 
+        # TODO: See Ingress: Has allow_partial etc.
         # Return an (args, kwargs) pair the respects the inner function's
         # argument kind restrictions.
         return self.inner_sig.args_and_kwargs_from_kwargs(func_kwargs)
 
+    # TODO: Use/test annotations of outer_sig
     @classmethod
     def from_signature(cls, inner_sig, outer_sig, _allow_reordering=False):
         """
+        A convienience ingress constructor to specify wrappings that affect arguments
+        independently.
 
-        :param inner_sig:
-        :param outer_sig:
-        :param _allow_reordering:
-        :return:
+        :param inner_sig: The signature of wrapped, inner function (or the inner
+        function itself)
+        :param outer_sig: The desired outer signature. Can also use a function (will
+        only take it's signature though).
+        :param _allow_reordering: Whether to allow ``outer_sig`` to reorder arguments.
+        :return: An ingress that will allow one to use a function having the
+        ``inner_sig`` signature to
 
         Say we wanted to get a version of the function:
 
@@ -923,24 +970,27 @@ class InnerMapIngress:
 
         That was equivalent to (note the kind, default and annotation differences):
 
-        >>> def g(w, x=1, y=2, z=10):
+        >>> def g(w, x=1, y: float = 2.0, z=10):
         ...     return w + x * y ** z
 
-
-        >>> ingress = InnerMapIngress.from_signature(
-        ... f, outer_sig=lambda w, x=1, y=2, z=10: None
-        ... )
-        >>> Sig(ingress)
-        <Sig (w, x=1, y=2, z=10)>
-        >>>
-        >>>
-        >>>
         >>> h = wrap(f, ingress=InnerMapIngress.from_signature(f, g))
+        >>> Sig(h)
+        <Sig (w, x=1, y: float = 2.0, z=10)>
+
+        Note we could have used ``...from_signature(Sig(f), Sig(g))`` as well,
+        since the method doesn't use the actual functions, just their signatures.
+
+        So we've seen that ``h`` takes on the signature (kind, defaults,
+        and annotations) of ``g``.
+        Let's see now that ``h`` actually computes, uses the defaults of ``g`` and
+        can doesn't have the position only restriction on ``w``.
+
         >>> assert h(0) == g(0) == 1024 == 0 + 1 * 2 ** 10
         >>> assert h(1,2) == g(1,2) == 2049 == 1 + 2 * 2 ** 10
         >>> assert h(1,2,3,4) == g(1,2,3,4) == 1 + 2 * 3 ** 4
         >>>
-        >>> assert h(w=1,x=2,y=3,z=4) == g(1,2,3,4) == 1 + 2 * 3 ** 4
+        >>> assert h(w=1,x=2,y=3,z=4) == g(1,2,3,4) == 1 + 2 * 3 ** 4  # w keyword arg!
+
         """
         outer_sig = Sig(outer_sig)
         return cls(
@@ -1065,6 +1115,10 @@ def wrap_from_sig(func, new_sig):
 # wrap tools
 
 
+def _only_keep_non_none_values(d: dict):
+    return {k: v for k, v in d.items() if v is not None}
+
+
 @double_up_as_factory
 def ch_names(func=None, **old_to_new_name):
     """Change the argument names of a function.
@@ -1084,32 +1138,34 @@ def ch_names(func=None, **old_to_new_name):
     >>> foo(alpha=1, b=2, gamma=3)
     7
     """
-    return Ingress.name_map(func, **old_to_new_name).wrap(func)
+    old_to_new_name = _only_keep_non_none_values(old_to_new_name)
+    if old_to_new_name:
+        return Ingress.name_map(func, **old_to_new_name).wrap(func)
+    else:
+        return func
 
 
 map_names = ch_names  # back-compatibility alias
 
+from i2.deco import _resolve_inclusion
 
-def include_exclude_ingress_factory(func, include=None, exclude=None):
-    """A pattern underlying any ingress that takes a subset of parameters (possibly
+
+# TODO: A more general version would allow include and exclude to be expressed as
+#     functions that apply to one or several properties of the params
+#     (name, kind, default, annotation).
+def include_exclude_ingress_factory(
+    func, include=None, exclude=None, *, allow_partial=False
+):
+    """
+    A pattern underlying any ingress that takes a subset of parameters (possibly
     reordering them).
 
     For example: Keep only required arguments, or reorder params to be able to
     partialize #3 (without having to partialize #1 and #2)
-
-    Note: A more general version would allow include and exclude to be expressed as
-    functions that apply to one or several properties of the params (name, kind, default,
-    annotation).
     """
     sig = Sig(func)
-    if isinstance(include, str):
-        include = include.split()
-    if isinstance(exclude, str):
-        exclude = exclude.split()
-    exclude = exclude or set()
-    include = [x for x in (include or sig.names) if x not in exclude]
-
-    return Ingress(inner_sig=sig, outer_sig=sig[include])
+    include = _resolve_inclusion(include, exclude, sig.names)
+    return Ingress(inner_sig=sig, outer_sig=sig[include], allow_partial=allow_partial)
 
 
 @double_up_as_factory
@@ -1135,8 +1191,16 @@ def include_exclude(func=None, *, include=None, exclude=None):
     return wrap(func, ingress=include_exclude_ingress_factory(func, include, exclude))
 
 
+# TODO: Not working completely with allow_removal_of_non_defaulted_params=True
+#  See https://github.com/i2mint/i2/issues/44
 @double_up_as_factory
-def rm_params(func=None, *, params_to_remove=()):
+def rm_params(
+    func=None,
+    *,
+    params_to_remove=(),
+    allow_removal_of_non_defaulted_params=False,
+    allow_partial=False,
+):
     """Get a function with some parameters removed.
 
     >>> from inspect import signature
@@ -1168,13 +1232,17 @@ def rm_params(func=None, *, params_to_remove=()):
     params_to_remove_that_do_not_have_defaults = set(params_to_remove) & set(
         sig.without_defaults.names
     )
-    assert not params_to_remove_that_do_not_have_defaults, (
-        f"Some of the params you want to remove don't have defaults: "
-        f'{params_to_remove_that_do_not_have_defaults}'
-    )
+    if not allow_removal_of_non_defaulted_params:
+        assert not params_to_remove_that_do_not_have_defaults, (
+            f"Some of the params you want to remove don't have defaults: "
+            f'{params_to_remove_that_do_not_have_defaults}'
+        )
 
     return wrap(
-        func, ingress=include_exclude_ingress_factory(func, exclude=params_to_remove)
+        func,
+        ingress=include_exclude_ingress_factory(
+            func, exclude=params_to_remove, allow_partial=allow_partial
+        ),
     )
 
 
@@ -1258,9 +1326,6 @@ def required_params_used_in_funcs(funcs):
     return {name for func in funcs for name in Sig(func).required_names}
 
 
-from i2.signatures import _call_forgivingly
-
-
 # TODO: Should we add some explicit/validation/strict options?
 def kwargs_trans(
     kwargs: dict = None, /, _recursive=False, _inplace=False, **key_and_val_func
@@ -1349,9 +1414,6 @@ def kwargs_trans(
 
 # ---------------------------------------------------------------------------------------
 
-from inspect import Parameter
-from dataclasses import make_dataclass
-
 empty = Parameter.empty
 
 
@@ -1375,15 +1437,7 @@ def kwargs_trans_to_extract_args_from_attrs(
 # TODO: kind lost here, only 3.10 offers dataclasses with some control over kind:
 #   See: https://stackoverflow.com/questions/49908182/how-to-make-keyword-only-fields-with-dataclasses
 def param_to_dataclass_field_tuple(param: Parameter):
-    t = (param.name, param.annotation, param.default)
-    if t[2] is empty:
-        t = t[:2]
-    if t[1] is empty:
-        if len(t) == 2:
-            t = t[0]
-        else:
-            t = (t[0], 'typing.Any', t[2])
-    return t
+    return param.name, param.annotation, param.default
 
 
 MethodFunc = Callable
@@ -1508,8 +1562,11 @@ def func_to_method_func(
     return method_func
 
 
+# TODO: See bind_funcs_object_attrs_old and get rid of it if bind_funcs_object_attrs
+#  is better and has all old has.
+# TODO: Make instances picklable (class is already, but not instances!)!!
 def bind_funcs_object_attrs(
-    funcs, init_params=(), *, cls=None,
+    funcs, init_params=(), *, cls=None, module=None, **extra_attrs
 ):
     """Transform one or several functions into a class that contains them as methods
     sourcing specific arguments from the instance's attributes.
@@ -1528,7 +1585,7 @@ def bind_funcs_object_attrs(
     >>> assert instance.foo(2, d='hello') == 'hello: 9' == foo(
     ...     a=1, b=2, c=3, d='hello')
     >>> str(signature(Klass))
-    "(a: 'typing.Any', c: 'typing.Any' = 2) -> None"
+    '(a, c=2) -> None'
     >>>
     >>> instance = Klass(a=1, c=3)
     >>> str(instance)
@@ -1563,6 +1620,123 @@ def bind_funcs_object_attrs(
     'd=5, e=2'
     """
 
+    if isinstance(init_params, str):
+        init_params = init_params.split()
+
+    dflt_cls_name = 'FuncsUnion'
+    if callable(funcs) and not isinstance(funcs, Iterable):
+        single_func = funcs
+        dflt_cls_name = camelize(getattr(single_func, '__name__', dflt_cls_name))
+        funcs = [single_func]
+
+    # If cls not given, make one from the funcs and init_params
+    if not isinstance(cls, type):
+        # if the class is not given, we need to make one
+        # ... for this we make a name
+        if isinstance(cls, str):
+            cls_name = cls
+        else:
+            cls_name = dflt_cls_name
+        # ... and an actual class
+        cls = _mk_base_class_for_funcs(cls_name, init_params, funcs)
+
+    for func in funcs:
+        method_func = func_to_method_func(func, init_params)
+        setattr(cls, method_func.__name__, method_func)
+
+    if module:
+        cls.__module__ = module
+    for attr_name, attr_val in extra_attrs.items():
+        setattr(cls, attr_name, attr_val)
+
+    return cls
+
+
+class PickleHelperMixin:
+    def __reduce__(self):
+        return self.__name__
+
+
+def _mk_base_class_for_funcs(cls_name, init_params, funcs):
+    """
+    Make a class with given init_params.
+
+    :param cls_name: The name the class should have
+    :param init_params: list of strings used to determine what arg names are in the init
+    :param funcs: list of functions used to add defaults and annotations to the init args
+    :return:
+    """
+    # Make the signature for the init
+    class_init_sig = Sig(init_params)
+    if funcs:
+        for func in funcs:
+            init_params_in_func_sig = list(
+                filter(None, map(Sig(func).get, init_params))
+            )
+            class_init_sig = class_init_sig.merge_with_sig(
+                init_params_in_func_sig,
+                default_conflict_method='fill_defaults_and_annotations',
+            )
+
+    dataclass_fields = list(map(param_to_dataclass_field_tuple, class_init_sig.params))
+    cls = make_dataclass(cls_name, dataclass_fields, bases=(PickleHelperMixin,))
+
+    return cls
+
+
+def bind_funcs_object_attrs_old(
+    funcs, init_params=(), *, cls=None,
+):
+    """Transform one or several functions into a class that contains them as methods
+    sourcing specific arguments from the instance's attributes.
+    >>> from inspect import signature
+    >>> from dataclasses import dataclass
+    >>>
+    >>> def foo(a, b, c=2, *, d='bar'):
+    ...     return f"{d}: {(a + b) * c}"
+    >>> foo(1, 2)
+    'bar: 6'
+    >>> Klass = bind_funcs_object_attrs_old(foo, init_params='a c')
+    >>> Klass.__name__
+    'Foo'
+    >>> instance = Klass(a=1, c=3)
+    >>> assert instance.foo(2, d='hello') == 'hello: 9' == foo(
+    ...     a=1, b=2, c=3, d='hello')
+    >>> str(signature(Klass))
+    '(a, c=2) -> None'
+    >>>
+    >>> instance = Klass(a=1, c=3)
+    >>> str(instance)
+    'Foo(a=1, c=3)'
+    >>> str(signature(instance.foo))
+    "(b, *, d='bar')"
+    >>> instance.foo(2, d='hello')
+    'hello: 9'
+    >>> instance.foo(10, d='goodbye')
+    'goodbye: 33'
+    >>> def foo(a, b, c):
+    ...     return a + b * c
+    ...
+    >>> def bar(d, e):
+    ...     return f"{d=}, {e=}"
+    ...
+    >>> @dataclass
+    ... class K:
+    ...     a: int
+    ...     e: int
+    ...
+    >>> C = bind_funcs_object_attrs([foo, bar], 'a e', cls=K)
+    >>> str(signature(C))
+    '(a: int, e: int) -> None'
+    >>> c = C(1,2)
+    >>> assert str(signature(c.foo)) == '(b, c)'
+    >>> c.foo(3,4)
+    13
+    >>> assert str(signature(c.bar)) == '(d)'
+    >>> c.bar(5)
+    'd=5, e=2'
+    """
+
     if isinstance(init_params, str):
         init_params = init_params.split()
 
@@ -1661,8 +1835,6 @@ def _mk_sig_from_params_and_funcs(params, funcs):
 # ---------------------------------------------------------------------------------------
 # Extended Wrapper class
 
-from i2 import call_forgivingly
-
 
 class WrapperValidationError(ValueError):
     """Raised when wrapper some construction params are not valid"""
@@ -2014,3 +2186,152 @@ def move_names_to_the_end(names, names_to_move_to_the_end):
         names_to_move_to_the_end = list(names_to_move_to_the_end)
     removed = [x for x in names if x not in names_to_move_to_the_end]
     return list(removed) + names_to_move_to_the_end
+
+
+# --------------------------------------------------------------------------------------
+# smart defaults
+
+# IDEA: Could make SmartDefault have an out and __call__, working like a meshed.FuncNode
+@dataclass
+class SmartDefault:
+    func_computing_default: Callable
+    original_default: Any = empty
+
+    def __repr__(self):
+        func = self.func_computing_default
+        func_name = getattr(func, '__name__', str(func))
+        if self.original_default is empty:
+            return f'SmartDefault({func_name})'
+        else:
+            dflt = self.original_default
+            if isinstance(dflt, str):
+                return f"SmartDefault({func_name}, '{dflt}')"
+            else:
+                return f'SmartDefault({func_name}, {dflt})'
+
+
+# PATTERN: Yet another "meshed dict completion"
+# IDEA: Could make SmartDefault have an out and __call__, working like a meshed.FuncNode
+def complete_dict_applying_functions(
+    d: dict, /, _only_if_name_missing=True, _allow_overwrites=False, **func_for_name
+):
+    """Complete dict ``d`` by applying function to variables in ``d``, sequentially.
+
+    That is, doing ``d[name] = func(**d)`` for all ``name, func in d.items()``.
+
+    Set ``_allow_overwrites=True`` to allow overwrites.
+
+    Set ``_only_if_name_missing=False`` to apply all functions of ``func_for_name``
+    regardless if the ``name`` already exists in ``d`` or not.
+
+    >>> func_for_name = dict(
+    ...     b=lambda a: a * 10, c=lambda a, b: a + b, d=lambda c: c * 2
+    ... )
+    >>> complete_dict_applying_functions(dict(a=1), **func_for_name)
+    {'a': 1, 'b': 10, 'c': 11, 'd': 22}
+
+    Notice that when ``b`` is present in input ``dict``, it's value is conserved.
+    That is, the ``b`` of ``func_for_name`` isn't applied to compute it.
+
+    >>> complete_dict_applying_functions(dict(a=1, b=2), **func_for_name)
+    {'a': 1, 'b': 2, 'c': 3, 'd': 6}
+
+    If you specify ``_only_if_name_missing=False``,
+    ``complete_dict_applying_functions`` will try to compute everything
+    ``func_for_name`` tells it too, regardless if the input dictionary contains the
+    key or not, resulting in an error:
+
+    >>> complete_dict_applying_functions(
+    ...     dict(a=1, b=2), **func_for_name, _only_if_name_missing=False
+    ... )
+    Traceback (most recent call last):
+      ...
+    i2.errors.OverwritesNotAllowed: You're not allowed to overwrite to the values of b
+
+    If you want, on the other hand, to allow overwrites, you can do so specifying
+    ``_allow_overwrites=True``:
+
+    >>> complete_dict_applying_functions(
+    ...     dict(a=1, b=2), **func_for_name,
+    ...     _only_if_name_missing=False, _allow_overwrites=True
+    ... )
+    {'a': 1, 'b': 10, 'c': 11, 'd': 22}
+
+    """
+
+    if _only_if_name_missing:
+        func_for_name = {
+            name: func for name, func in func_for_name.items() if name not in d
+        }
+    elif not _allow_overwrites and not func_for_name.keys().isdisjoint(d):
+        raise OverwritesNotAllowed.for_keys(func_for_name.keys() & d)
+
+    for name, func in func_for_name.items():
+        # TODO: Can optimize using under-the-hood of call_forgivingly
+        #  (extracting from _kwargs directly)
+        # complete the keyword arguments with defaults computed from existing arguments
+        d[name] = call_forgivingly(func, **d)
+
+    return d
+
+
+def _compute_new_sigs(func, /, **smart_defaults):
+    original_sig = Sig(func)
+    return original_sig.ch_defaults(
+        **{
+            name: SmartDefault(
+                func_computing_default=func,
+                original_default=original_sig.defaults.get(name, empty),
+            )
+            for name, func in smart_defaults.items()
+        }
+    )
+
+
+@double_up_as_factory
+def add_smart_defaults(
+    func=None, *, _only_if_name_missing=True, _allow_overwrites=False, **smart_defaults,
+):
+    """Add smart defaults to function.
+
+    Smart defaults compute defaults of inputs based on the other given inputs.
+
+    :param func: The function to wrap.
+    :param smart_defaults: ``input_arg=function_to_compute_it_from_other_args`` where
+        the function's argument names must match the argument names of ``func``, the
+        wrapped function.
+    :return:
+
+    >>> def xyz_sum(x, y, z='c'):  return x + y + z
+    >>> def times_two(x):  return x * 2
+    >>> def just_z():  return 'Z'
+    >>> f = add_smart_defaults(xyz_sum, y=times_two, z=just_z)
+    >>>
+    >>> f('a', 'b', 'c')
+    'abc'
+    >>> f('a', 'b')
+    'abZ'
+    >>> f('a')
+    'aaaZ'
+    >>> f
+    <i2.Wrap xyz_sum(x, y=SmartDefault(times_two), z=SmartDefault(just_z, 'c'))>
+
+    """
+    names_not_in_func_arguments = smart_defaults.keys() - Sig(func).names
+    assert (
+        not names_not_in_func_arguments
+    ), f"These weren't argument names of the {func} function: {names_not_in_func_arguments}"
+    kwargs_trans = partial(
+        complete_dict_applying_functions,
+        _only_if_name_missing=_only_if_name_missing,
+        _allow_overwrites=_allow_overwrites,
+        **smart_defaults,
+    )
+    smart_defaults_ingress = Ingress(
+        inner_sig=Sig(func),
+        kwargs_trans=kwargs_trans,
+        outer_sig=_compute_new_sigs(func, **smart_defaults),
+        apply_defaults=False,
+    )
+
+    return smart_defaults_ingress.wrap(func)
